##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.auth.ds8k.base import Auth as AuthBase
from pyds8k.resources.ds8k.v1.common.base import Base

AUTH_URL = '/tokens'
base_url = Base.base_url


class Auth(AuthBase):

    base_url = Base.base_url



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
IBM DS8K CLI Client Module

:module: pyds8k

"""
PYDS8K_DEFAULT_LOGGER = "pyds8k"

version_tuple = (1, 2, 0)


def get_version_string():
    """Current version of Client."""
    if isinstance(version_tuple[-1], str):
        return '.'.join(map(str, version_tuple[:-1])) + '.' + version_tuple[-1]
    return '.'.join(map(str, version_tuple))


version = get_version_string()



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
advanced PPRC interface.
"""
from pyds8k.base import ManagerMeta, ResourceMeta
from ..common.base import Base, ReadOnlyManager
from ..common.types import DS8K_CS_PPRC
from ..volumes import Volume, VolumeManager
from ..systems import System, SystemManager


class PPRC(Base, metaclass=ResourceMeta):
    resource_type = DS8K_CS_PPRC

    _template = {'id': '',
                 'type': '',
                 'state': '',
                 'source_system': '',
                 'target_system': '',
                 'source_volume': '',
                 'target_volume': '',
                 }

    related_resource = {'_source_volume': (Volume, VolumeManager),
                        '_source_system': (System, SystemManager),
                        '_target_volume': (Volume, VolumeManager),
                        '_target_system': (System, SystemManager),
                        }

    def _update_volume_info(self, info):
        # Handle for bug in DS8000 RESTful API /api/v1/cs/pprcs:
        # When it responds, source_volume and target_volume use
        # "name" as the key of id field.
        for key in ['source_volume', 'target_volume']:
            if 'name' in info[key] and 'id' not in info[key]:
                info[key]['id'] = info[key].pop('name')
        return info

    def _add_details(self, info, force=False):
        self._start_updating()
        self._update_volume_info(info)
        self._stop_updating()
        super(PPRC, self)._add_details(info, force=force)


class PPRCManager(ReadOnlyManager, metaclass=ManagerMeta):
    """
    Manage advanced PPRC resources.
    """
    resource_class = PPRC
    resource_type = DS8K_CS_PPRC



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Host interface.
"""
from pyds8k.base import ManagerMeta, ResourceMeta
from .common.base import Base, BaseManager
from .common.mixins import FCPortMixin, HostPortMixin, VolumeMixin, VolmapMixin
from .common import types


class Host(FCPortMixin, HostPortMixin, VolumeMixin, VolmapMixin, Base,
           metaclass=ResourceMeta):
    resource_type = types.DS8K_HOST
    id_field = 'name'
    alias = {'id': 'host_id'}
    _template = {'name': '',
                 'state': '',
                 'hosttype': '',
                 'addrmode': '',
                 'addrdiscovery': '',
                 'lbs': '',
                 types.DS8K_VOLUME: '',
                 types.DS8K_IOPORT: '',
                 types.DS8K_HOST_PORT: '',
                 }
    readonly_fileds = ('state', 'addrmode', 'addrdiscovery', 'lbs',
                       types.DS8K_HOST_PORT,
                       )
    related_resources_collection = (types.DS8K_VOLUME,
                                    types.DS8K_IOPORT,
                                    types.DS8K_HOST_PORT,
                                    types.DS8K_VOLMAP,
                                    )
    # def __repr__(self):
    #    return "<Host: {0}>".format(self.id)

    def update_host_add_ioports(self, port_ids=[]):
        if not port_ids:
            return self.update_host_add_ioports_all(self.id)
        updated_port_ids = self._update_ioports_and_return_ids(port_ids)

        _, res = self.one(types.DS8K_HOST,
                          self.id,
                          rebuild_url=True
                          ).update({'ioports': updated_port_ids})
        return res

    def update_host_rm_ioports(self, port_ids=[]):
        if not port_ids:
            return self.update_host_rm_ioports_all(self.id)
        updated_port_ids = self._update_ioports_and_return_ids(port_ids, '-')

        _, res = self.one(types.DS8K_HOST,
                          self.id,
                          rebuild_url=True
                          ).update({'ioports': updated_port_ids})
        return res

    def _update_ioports_and_return_ids(self, port_ids, operator='+'):
        ports = []
        if not isinstance(port_ids, list):
            port_ids = [port_ids, ]
        for port_id in port_ids:
            port = port_id
            if not isinstance(port_id, Base):
                port = self.one(types.DS8K_IOPORT, port_id)
            ports.append(port)
        updated = self._update_list_field(types.DS8K_IOPORT, ports, operator)
        return [port.id for port in updated]


class HostManager(BaseManager, metaclass=ManagerMeta):
    """
    Manage Host resources.
    """
    resource_class = Host
    resource_type = types.DS8K_HOST

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id,
                         url=url, obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from . import ioports, flashcopy, events, mappings, pprc, eserep, \
    users, systems, nodes, marrays, encryption_groups, io_enclosures, \
    pools, tserep, lss, volumes, host_ports, hosts
from .cs import pprcs, flashcopies

__all__ = (
    'ioports',
    'flashcopy',
    'flashcopies',
    'events',
    'mappings',
    'pprc',
    'eserep',
    'pprcs',
    'users',
    'systems',
    'nodes',
    'marrays',
    'encryption_groups',
    'io_enclosures',
    'pools',
    'tserep',
    'lss',
    'volumes',
    'host_ports',
    'hosts'
)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Storage volume interface.
"""
from pyds8k.base import ManagerMeta, ResourceMeta
from .common.base import Base, BaseManager
from .common import types
from .pools import Pool, PoolManager
from .lss import LSS, LSSManager
from .hosts import Host, HostManager


class Volume(Base, metaclass=ResourceMeta):
    resource_type = types.DS8K_VOLUME
    # id_field = 'id'

    # Set the value to None if the field is not required when creation.
    _template = {'id': None,
                 'name': '',
                 'state': '',
                 'cap': '',
                 'real_cap': None,
                 'virtual_cap': None,
                 'captype': '',
                 'stgtype': '',
                 'VOLSER': None,
                 'allocmethod': '',
                 'tp': '',
                 'capalloc': '',
                 'MTM': None,
                 'datatype': '',
                 'easytier': '',
                 'tieralloc': [],
                 'lss': '',
                 'pool': '',
                 types.DS8K_HOST: None,
                 types.DS8K_FLASHCOPY: None,
                 types.DS8K_PPRC: None,
                 }
    fb_template = _template.copy()
    fb_template.update({'stgtype': types.DS8K_VOLUME_TYPE_FB})

    ckd_template = _template.copy()
    ckd_template.update({'stgtype': types.DS8K_VOLUME_TYPE_FB})

    template_dict = {types.DS8K_VOLUME_TYPE_FB: fb_template,
                     types.DS8K_VOLUME_TYPE_CKD: ckd_template,
                     }
    readonly_fileds = ('state', 'stgtype', 'VOLSER', 'allocmethod', 'tp',
                       'capalloc', 'MTM', 'datatype', 'easytier', 'tieralloc',
                       'lss'
                       )
    related_resource = {'_pool': (Pool, PoolManager),
                        '_lss': (LSS, LSSManager)
                        }
    related_resources_collection = (types.DS8K_HOST,
                                    types.DS8K_FLASHCOPY,
                                    types.DS8K_PPRC)

    def __init__(self, client, manager=None, url='', info={},
                 resource_id=None,
                 parent=None,
                 loaded=False,
                 volume_type=types.DS8K_VOLUME_TYPE_FB,
                 ):
        super(Volume, self).__init__(client,
                                     manager=manager,
                                     url=url,
                                     info=info,
                                     resource_id=resource_id,
                                     parent=parent,
                                     loaded=loaded,
                                     )
        self.volume_type = volume_type
        self._verify_volume_type()
        self._template = self.get_template_from_volume_type(volume_type)

    def __repr__(self):
        return "<Storage Volume: {0}>".format(self._get_id())

    def get_template_from_volume_type(self, volume_type):
        return self.template_dict[volume_type]

    def _verify_volume_type(self):
        self._verify_type(self.volume_type, types.DS8K_VOLUME_TYPES)

    def _set_hosts(self):
        host_list = self.representation.get('host', [])
        if host_list:
            host_obj_list = []
            for host in host_list:
                host_obj = Host(self.client,
                                manager=HostManager(self.client),
                                info=host,
                                loaded=False,
                                )
                host_obj_list.append(host_obj)
            self.representation[types.DS8K_HOST] = [
                h.host_id for h in host_obj_list]
            setattr(self, types.DS8K_HOST, host_obj_list)

    def _add_details(self, info, force=False):
        super(Volume, self)._add_details(info, force=force)
        self._start_updating()
        self._set_hosts()
        self._stop_updating()


class VolumeManager(BaseManager, metaclass=ManagerMeta):
    """
    Manage Storage Volume resources.
    """
    resource_class = Volume
    resource_type = types.DS8K_VOLUME

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id, url=url,
                         obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    # def post(self, url='', body=None):
    #    return self._post(url, body)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
advanced FlashCopies interface.
"""
from pyds8k.base import ManagerMeta, ResourceMeta
from ..common.base import Base, BaseManager
from ..common.types import DS8K_CS_FLASHCOPY, DS8K_FLASHCOPY
from ..volumes import Volume, VolumeManager


class FlashCopy(Base, metaclass=ResourceMeta):
    resource_type = DS8K_CS_FLASHCOPY
    _template = {'id': None,
                 'persistent': None,
                 'recording': None,
                 'backgroundcopy': None,
                 'state': None,
                 'options': [],
                 'volume_pairs': []
                 }

    related_resource = {'_volume_pairs': [{
        'source_volume': (Volume, VolumeManager),
        'target_volume': (Volume, VolumeManager)
    }]
    }

    def __repr__(self):
        return "<FlashCopy: {0}>".format(self._get_id())

    def _add_details(self, info, force=False):
        super(FlashCopy, self)._add_details(info, force=force)
        if DS8K_FLASHCOPY in info:
            self._id = info[DS8K_FLASHCOPY][0]['id']


class FlashCopyManager(BaseManager, metaclass=ManagerMeta):
    """
    Manage advanced FlashCopies resources.
    """
    resource_class = FlashCopy
    resource_type = DS8K_CS_FLASHCOPY

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id, url=url,
                         obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)
